:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
@echo off

    :: Porting By Francesco Poscetti aka einstein1969, original idea by Guido D'Albore

	:: Use raster font 8x8 for better aspect ratio, save in utf8

	:: v 0.1.3	07/05/2023
	::
	::		- rotation is function of the time.
	::		- new autotune algorithm
	::
	:: v 0.1.1	30/04/2023 
	::
	::		- implemented 2D clipping
	::		- Decreased radius
	::
	:: v 0.1.0  	29/04/2023 Bigger sphere point/Point Decoration, better autotune algorithm
	::
	:: v 0.0.3 	update sin code, use vt100, , use bullet ·
	::		code optimization for faster execution, 
	:: 		auto tune number of points for dinamic resolution based on pc performance (static)
	::		better zbuffer, break near infinite loop by jeb.
	::
	:: v 0.0.2 	passed from 16.6 FPS to ~20.0 FPS

    :: Variables enigma:

    :: _R   =   Sphere radius
    :: rt   =   Rotation angle
    :: ds   =   Distance Z of sphere
    :: _dx,_dy   =   Offset x,y of projecting coordinates in 2D plane.
    :: sx,cx   =   Sine, cosine of rotation angle
    :: a,b,c   =   Rotation coordinate 3D
    :: a,b   =   X,Y coordinate of projecting 3D to 2D
    :: c   =   Z coordinate

    :: Ref:
    :: https://www.dostips.com/forum/viewtopic.php?f=3&t=5594
    :: https://invisible-island.net/xterm/ctlseqs/ctlseqs.html
    :: https://learn.microsoft.com/en-us/windows/console/console-virtual-terminal-sequences

:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
  title Sphere 3D
  set /a _cols=120, _lines=90
  mode %_cols%,%_lines%
  set /a _cols-=2, _lines-=2

  rem utf 8
  chcp 65001 >nul
  for /F %%a in ('Echo(prompt $E^| cmd') Do Set _ESC=%%a

:ChoiceLabel
  echo(
  echo Select the maximum speed. Choose a number between 0 and 9 where:
  echo(
  echo 0 = 100.0 FPS (Frame Per Second) Very Fast, few sphere point
  echo 1 = 50.0 FPS 
  echo 2 = 33.3 FPS
  echo 3 = 25.0 FPS 
  echo 4 = 20.0 FPS
  echo 5 = 16.6 FPS
  echo 6 = 14.2 FPS
  echo 7 = 12.5 FPS
  echo 8 = 11.1 FPS
  echo 9 = 10.0 FPS Slow, high number of sphere point
  echo(

  choice /M "Select the maximum speed. Choose a number between 1 and 9 where 1 very fast and 9 slow. Default [3] ?" /D 3 /C 0123456789 /T 6 /N
  if %errorlevel% gtr 0 (set /a "_FPS=100/(%errorlevel%)") else goto :ChoiceLabel

(  
  for /F "Tokens=1 delims==" %%v in ('set') do set "%%v="
  set /a "_cols=%_cols%, _lines=%_lines%"
  set "_ESC=%_ESC%"
  set "_FPS=%_FPS%"
)

  setlocal EnableDelayedExpansion

  :: Big loop 2^81 = 2.417.851.639.229.258.349.412.352 with very fast break
  set "_loop=FOR /L %%# in (1,1,2) do "
  %_loop%%_loop% set "_loop=!_loop!!_loop!!_loop!"

  rem https://www.dostips.com/forum/viewtopic.php?f=3&t=4896
  set "_SIN(x)=(a=((x)%%62832)+(((x)%%62832)>>31&62832), b=(a-15708^a-47124)>>31,a=(-a&b)+(a&~b)+(31416&b)+(-62832&(47123-a>>31)),a-a*a/1875*a/320000+a*a/1875*a/15625*a/16000*a/2560000-a*a/1875*a/15360*a/15625*a/15625*a/16000*a/44800000)"

:: Main loop

  :: Hide the cursor
  <nul set /p "=%_ESC%[?25l"

  :: ndx min=8 - max=42 for var "s"
  set /a "_Radius=150, _dx=_cols/2, _dy=_lines/2, _dz=4000, ndx=10, ds=0, rt=0, i=1"

  %= *Φ•○♥øo█ =%
  set "_DEC=♥"

  rem remove this rem for bigger sphere point. This can affect the correct functioning of 2D clipping!
  rem set "_DEC=█%_ESC%[A█\%_ESC%[D%_ESC%[B█%_ESC%[B%_ESC%[2D█/%_ESC%[3D\%_ESC%[2A%_ESC%[D/%_ESC%[B○"

	%=				Calculate a sphere point					 Rotation      																																			      Projection of 3D to 2D				=%
  set "_formula=a=%_Radius%*%%h/10000*%%u/10000, c=%_Radius%*%%g/10000*%%u/10000, b=%_Radius%*%%v/10000, a=(a*%%b/10000-(b*%%a+c*%%b)/10000*%%a/10000)*%%b/10000-(b*%%b-c*%%a)/10000*%%a/10000, b=(%_Radius%*%%h/10000*%%u/10000*%%b/10000-(b*%%a+c*%%b)/10000*%%a/10000)*%%a/10000+(b*%%b-c*%%a)/10000*%%b/10000, c=%_Radius%*%%h/10000*%%u/10000*%%a/10000+(%_Radius%*%%v/10000*%%a+c*%%b)/10000*%%b/10000, a=ds*a/(c-%_dz%)+%_dx%, b=ds*b/(c-%_dz%)+%_dy%"

  for /f "tokens=1-4 delims=:.," %%a in ("!time: =0!") do set /a "t1=(((1%%a*60)+1%%b)*60+1%%c)*100+1%%d-36610100, t3=t1, tuning=1"

  For /L %%L in () do (	

	%= Prepare for faster dinamic execution =%

	set /a "_ndy=ndx/2, _stepx=62832/ndx, _stepy=31416/_ndy"

	%= rainbow colors plus meridians=%
	set /a _H=0, _dH=359/ndx, _V=10000, angle=0
	set "_X2="

	for /L %%C in (1,1,!ndx!) do (

	  set /a "_mm = (_H*10000/60 %% 20000) - 10000, _mm=(_mm>>31|1)*_mm, _C = _V, _X = _C *(10000 - _mm)/10000, _C=_C*255/10000, _X=_X*255/10000"
	  if !_H! lss 60 (set /a _R=_C, _G=_X, _B=0) else (
   		if !_H! lss 120 (set /a _R=_X, _G=_C, _B=0) else (
			if !_H! lss 180 (set /a _R=0, _G=_C, _B=_X) else (
				if !_H! lss 240 (set /a _R=0, _G=_X, _B=_C) else (
					if !_H! lss 300 (set /a _R=_X, _G=0, _B=_C) else (
						if !_H! lss 360 (set /a _R=_C, _G=0, _B=_X) else (echo HSV ERROR!)
	  )))))
	  %= compute for 216 color map =%
          set /A "_color=((_R-35)/40+((_R-48)>>8)-((_R-75)>>8))*36 + ((_G-35)/40+((_G-48)>>8)-((_G-75)>>8))*6 + ((_B-35)/40+((_B-48)>>8)-((_B-75)>>8)) + 16, _H+=_dH"

	  %= meridians =%
	  set /A "sx=%_SIN(x):x=angle%, cx=%_SIN(x):x=(15708-angle)%, angle+=_stepx"

	  set "_X2=!_X2!"!sx! !cx! !_color! " "
	)
	set "_X=!_X2!"

	%= parallels =%
	set /a "angle=_stepy"
	set "_Y="
	for /L %%X in (2,1,!_ndy!) do (	
		set /A "sx=%_SIN(x):x=angle%, cx=%_SIN(x):x=(15708-angle)%, angle+=_stepy"
		set "_Y=!_Y!"!sx! !cx!" "	
	)
	set "angle="

	%= Poles 0-pi =%
	set /A "sx=%_SIN(x):x=0%, cx=%_SIN(x):x=(15708)%"
	set "_PX="!sx! !cx!""
	set "_PY="!sx! !cx!" "
	set /A "sx=%_SIN(x):x=31416%, cx=%_SIN(x):x=(15708-31416)%"
	set "_PY=!_PY!"!sx! !cx!""

	%= Macros =%
	rem if ndx gtr 16 then split string else not split. keep separate to overcome 8k limit
	if !ndx! gtr 16 (
		rem init1
		set "_init1=for /L %%G in (1,1,^!ndx^!) do set "r[%%G]=^^!r[%%G]^^!%_ESC%[30m""
		rem init2
		set "_init2=set "t[1]=^^!t[1]^^!%_ESC%[91m""
		rem init3
		set "_init3=For %%C in (^!z^!) do set "t[%%C]=^^!t[%%C]^^!%_ESC%[38;5;%%im""
		rem acc1
	  	set "_Acc1=for %%G in (^!z^!) do ( set "t[%%G]=^^!t[%%G]^^!^^!d^^!" & set "r[%%G]=^^!r[%%G]^^!^^!d^^!" )"
		rem echo1
		set "_echo1=for /L %%G in (1,1,^!ndx^!) do echo %_ESC%M^!t[%%G]^!"
		rem divide1
		%= keep space for empty "t" =%
		set "_div1=for /L %%G in (1,1,^!ndx^!) do ( set "t[%%G]=^^!r[%%G]^^! " & set "r[%%G]= " )"
		rem free1
		set "_free1=for /L %%G in (1,1,^!ndx^!) do set "t[%%G]=^^!t[%%G]:* =^^!""
	) else (
		rem init1
		set "_init1=set "r=^^!r^^!%_ESC%[30m""
		rem init2
		set "_init2=set "t=^^!t^^!%_ESC%[91m""
		rem init3
		set "_init3=set "t=^^!t^^!%_ESC%[38;5;%%im""
		rem acc1
	  	set "_Acc1=set "t=^^!t^^!^^!d^^!" & set "r=^^!r^^!^^!d^^!""
		rem echo1
		set "_echo1=echo ^!t^!"
		rem divide1
		%= keep space for empty "t" =%
		set "_div1=set "t=^^!r^^! " & set "r= ""
		rem free1
		set "_free1=set "t=^^!t:* =^^!""
	)

	call :rotate dir
	
	set /a ndx+=dir

  )

goto :eof

:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
:rotate

setlocal

(

  for /F "Tokens=1 delims==" %%v in ('set _') do set "%%v="

  %= clear screen and go cursor in 1,1-Home =%
  set/p "=%_ESC%[2J%_ESC%[H"<nul

  %_loop% (

	for /f "tokens=1-4 delims=:.," %%a in ("!time: =0!") do set /a "t2=(((1%%a*60)+1%%b)*60+1%%c)*100+1%%d-36610100, tD=t2-t1, tD+=(tD>>31) & 8640000"

	%= Full loop rt=(120)*31416/60 or multiple of 120. 120*31416/60/100/5=126 is in 5 seconds=%
	set /a "rt=tD*79, fc+=1, tD2=t2-t3, tD2+=(tD2>>31) & 8640000"

	%= every 2 seconds, don't worry if dos use 64Hz clock =%
	if !tD2! gtr 200 (

		set /a "FPS=fc*100*10/tD2, dfps=(FPS/10-%_FPS%), te=-%_FPS%/2"

		if !tuning! equ 1 (
			title Sphere 3D - Tuning FPS ... FPS=[!FPS:~0,-1!.!FPS:~-1!]
			set /a "ndx=6, dir=0, tuning=2, t3=t2, fc=0"
			goto :exit_rotate
		) else 	(
			if !tuning! equ 2 (
				if !dfps! gtr 0 (
					title Sphere 3D - Tuning FPS ... FPS=[!FPS:~0,-1!.!FPS:~-1!]
					set /a "dir=2, tuning=2, t3=t2, fc=0"
					if %ndx% lss 42 goto :exit_rotate
				) else (
					set /a "tuning=3, wait=t3+1000"
				)
			) else (
				%= tuning=3 =%
				if !dfps! lss !te! (
					set /a "tuning=1
				) else (
					set /a "te=t3-wait"
					if !te! gtr 0 set /a "tuning=2"
				)
			)
		)
 
		title Sphere 3D - FPS=[!FPS:~0,-1!.!FPS:~-1!] max=[%_FPS%] - number of meridian=[%ndx%]
		set /a "fc=0, t3=t2"
		set "FPS="
		set "te="
		set "dir="
		set "dfps="
	)
	set "tD=" & set "tD2="

	%= sphere distance =%
	if !ds! lss 1000 (set /a ds+=5)

	%= Cos(x)=Sin(PI/2-x) =%
	set /A "sx=%_SIN(x):x=rt%, cx=%_SIN(x):x=(15708-rt)%"

	For /f "tokens=1,2" %%a in ("!sx! !cx!") do (
		set "cx=" & set "sx="

		set "s=!s!%_ESC%[92m"
		%_init1%	%= set "r=!r!%_ESC%[30m" =%
		%_init2%	%= set "t=!t!%_ESC%[91m" =%

		%= Poles =%
        	for %%f in (%_PX%) do for /f "tokens=1,2" %%g in (%%f) do (
			for %%t in (%_PY%) do for /f "tokens=1,2" %%u in (%%t) do (

				set /a "%_formula%"

				%= 2D Clipping =%
				if !a! gtr 2 if !b! gtr 2 if !a! lss %_cols% if !b! lss %_lines% (
					if !c! lss 0 (
						set "s=!s!%_ESC%[!b!;!a!H·"
					) else (
						set "d=!d!%_ESC%[!b!;!a!H%_DEC%"
				)
		)))

		set /A "z=1"

		%= meridians plus colors=%
		for %%f in (%_X%) do for /f "tokens=1,2,3" %%g in (%%f) do (

	  		%_init3%

	  		%= parallels =%
          		for %%t in (%_Y%) do for /f "tokens=1,2" %%u in (%%t) do (

	    			set /a "%_formula%"

	    			%= 2D Clipping =%
	    			if !a! gtr 2 if !b! gtr 2 if !a! lss %_cols% if !b! lss %_lines% (
	    				if !c! lss 0 (
						set "s=!s!%_ESC%[!b!;!a!H·"
	    				) else (
						set "d=!d!%_ESC%[!b!;!a!H%_DEC%"
	    				)
	    			)

          		) %= parallels =%

	  		%_Acc1%

	  		set "d="

	  		set /A "z+=1"
  		
		) %= meridians =%

	) %= END OF For /f "tokens=1,2" %%a in ("!sx! !cx!") do =%

	%= intrinsic cls =%
	%= echo zbuffer like =%
	echo(!s!
	%_echo1%

	%_div1%

	%= Foreground Black plus space =%
	set "s=!s:92m=30m! "
	
	if !i! lss 3 ( set /a "i+=1" ) else (
		set "s=!s:* =!"
		%_free1%
	)
  ) %= end_loop =%
)
:exit_rotate

(endlocal & set /a ds=%ds%, %1=%dir%, t1=%t1%, t3=%t3%, ndx=%ndx%, tuning=%tuning%)

goto :eof
:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
